; RUNFROMMENU EQU 1

; Tell the AS assembler we want to code for NGPC
;
        cpu 93C141
        maxmode on
        include "stddef96.inc"


        include "hardware.inc"  ; Include hardware defines
        include "system.inc"    ; Include System Call defines
        
;
; Define variables
;

THISLEVEL  EQU _MAINRAM ; editable level storage

START_OF_RAM    EVAL    _MAINRAM + 400h
        RAMDB VBCounter
        RAMDB VBTempCounter
        RAMDB VBStartTimer
        RAMDB LevelWidth
        RAMDB LevelHeight
        RAMDB PlayerY
        RAMDB PlayerX
        RAMDB ControlAnim
        RAMDB PreviousMove
        RAMDB NextMove
        RAMDB NextMoveOK
        RAMDB PlayerMoved
        RAMDB PlayerPushing
        RAMDB Lives
        RAMDB DiamondsRemaining
        RAMDB TimeRemaining
        RAMDD Score
        RAMDB LevelNumber
        RAMDB AnimY
        RAMDB AnimX
        RAMDB DisableInput
        RAMDB CountdownToLevelReset
        
        RAMDB VisX
        RAMDB VisY
        RAMDB Scroll
        RAMDB ScrollSizeX ; this is levelwidth - 10 (screen size)
        RAMDB ScrollSizeY ; this is levelheight - 10
        RAMDB InViewX ; set if when doing gravity the rock/diamond is in view
        RAMDB InViewY
        RAMDB TopViewPos
        RAMDB BottomViewPos

        RAMDB FireFlyAnimNo
        RAMDB ExitAnimNo
        RAMDB ButterFlyAnimNo
        RAMDB OnToExit
        RAMDD LevelName
        RAMDD tmpLevelNumberStr

    
TILEBLANK      EQU 0    
TILESOIL       EQU 1
TILEWALL       EQU 2
TILEROCK       EQU 3
TILEDIAMOND    EQU 4
TILEEXIT       EQU 5
TILESOLIDWALL  EQU 6
TILEFIREFLY    EQU 7
TILEBUTTERFLY  EQU 8
TILEEXPLOSION  EQU 9 ; 10, 11 and 12

TILEPLAYER     EQU 52
PLAYERSPRNO    EQU 0

PLAYERBASE  EQU 100
PLSTAND  EQU   PLAYERBASE + 2
PLRIGHT  EQU   PLAYERBASE + 12
PLLEFT   EQU   PLAYERBASE + 16
PLUP     EQU   PLAYERBASE + 8
PLDOWN   EQU   PLAYERBASE + 4
PLEXIT   EQU   PLAYERBASE + 24


FIREFLYRIGHT   EQU 000h
FIREFLYDOWN    EQU 040h
FIREFLYLEFT    EQU 080h
FIREFLYUP      EQU 0C0h
TILEFIREFLYMOVED   EQU 020h + TILEFIREFLY

BUTTERFLYRIGHT   EQU 000h
BUTTERFLYDOWN    EQU 040h
BUTTERFLYLEFT    EQU 080h
BUTTERFLYUP      EQU 0C0h
TILEBUTTERFLYMOVED   EQU 020h + TILEBUTTERFLY



SPACERESERVE     EQU 020h
TILEROCKPUSHED   EQU 021h

TILEFALLINGLEFT      EQU 080h
TILEFALLINGRIGHT     EQU 040h
TILEFALLINGDOWN      EQU 0C0h

; status bar stuff
STATUSCHAR     EQU (STATUSSTART - CHARSET) / 16
ST_DIAMOND     EQU STATUSCHAR
ST_TIME        EQU STATUSCHAR + 1
ST_LIFE        EQU STATUSCHAR + 2
ST_NUMBERS     EQU STATUSCHAR + 3

MAXLEVELWIDTH  EQU 32
MAXLEVELWIDTHFASTMUL EQU 5 ; used for multiplying by maxlevelwidth (sll 5, x)


   IFDEF RUNFROMMENU
HEADER_OFS      EQU     0272000H
   ELSE
HEADER_OFS      EQU     0200000H
   ENDIF

;
; Standard cartridge header
;
    org    HEADER_OFS
    db     " LICENSED BY SNK CORPORATION"    ; 28 bytes license string
    dd     _start                            ; Program Counter
    dw     0                                 ; Catalog number
    db     0                                 ; Sub catalog number
    db     10h                               ; colour or b+w (10h = colour)
;           123456789012
    db     "DIAMOND RUN "         ; Game name (12 bytes)
    db     3,0,0,0                           ; WorkInProgress tab of Menu      
    dd     0,0,0                             ; padding - reserve for future use

; User Interrupt Vectors

UserIntVect:
        dd      nada            ; Software Interrupt (SWI 3)
        dd      nada            ; Software Interrupt (SWI 4)
        dd      nada            ; Software Interrupt (SWI 5)
        dd      nada            ; Software Interrupt (SWI 6)
        dd      nada            ; RTC Alarm Interrupt
        dd      VBlankInt       ; Vertical Blanking Interrupt
        dd      nada            ; Interrupt from Z80
        dd      nada            ; Timer Interrupt (8 bit timer 0)
        dd      nada            ; Timer Interrupt (8 bit timer 1)
        dd      nada            ; Timer Interrupt (8 bit timer 2)
        dd      nada            ; Timer Interrupt (8 bit timer 3)
        dd      nada            ; Serial Transmission Interrupt
        dd      nada            ; Serial Reception Interrupt
        dd      nada            ; (Reserved)
        dd      nada            ; End Micro DMA Int (MicroDMA 0)
        dd      nada            ; End Micro DMA Int (MicroDMA 1)
        dd      nada            ; End Micro DMA Int (MicroDMA 2)
        dd      nada            ; End Micro DMA Int (MicroDMA 3)

nada:   reti

; *** Start of User Code ***

_start:
        calr    OS_VERSION      ; Initialize NGP or NGPC mode

        set     6,(rUSERA)      ; User Answer

; Install User Interrupt Vectors

        lda     xix,(UserIntVect)
        lda     xiy,(rSWI3)
        ld      b,18
UIVloop:
        ld      xwa,(xix+)
        ld      (xiy+),xwa
        djnz    b,UIVloop

; set up screen size 160x152
    ld   (08002h),0
    ld   (08003h),0
    ld   (08004h),0a0h
    ld   (08005h),98h

    ldw (_BGCPAL), 0h
    ld (8118h), 80h

; enable interrupts
        ei


_gamestart:
    calr DoTitleScreen

    calr InitialiseMemory

; copy charset to character RAM
        ld bc, (ENDCHARSET - CHARSET)/2
        ld xde, _TILERAM
        ld xhl, CHARSET
        ldirw (xde+),(xhl+)

; copy palette 1 to character palette RAM
        ld bc, (ENDPALETTE1-PALETTE1)/2
        ld xde, _SCR1PAL
        ld xhl, PALETTE1
        ldirw (xde+),(xhl+)

; copy palette 2 to character palette RAM
        ld bc, (ENDPALETTE2-PALETTE2)/2
        ld xde, _SCR2PAL
        ld xhl, PALETTE2
        ldirw (xde+),(xhl+)

; copy sprite palette to palette RAM
       ld bc, (ENDSPRPALETTE-SPRPALETTE)/2
       ld xde, _SPRPAL
       ld xhl, SPRPALETTE
       ldirw (xde+),(xhl+)

        ; start of game
        ld (Lives), 3
        ld (LevelNumber), 1

_levelrestart:
	calr StopTimer

        ; replace animated exit
        ld xhl, NORMALEXIT
        ld bc, 32 ; 4 chars
        ld xde, _TILERAM + (TILEEXIT * 64)
        ldirw (xde+),(xhl+)

        ; init variables
        ld a, 0
        ld (VisX), a
        ld (VisY), a
        ld (DisableInput), 0
        ld (CountdownToLevelReset), 0
        ld (OnToExit), 0
        ld (NextMoveOK), 0
        ld (NextMove), 0

        cp (Lives), 0
        jr nz, _gameon
	
	calr DoDisplayGameOver
	jp _gamestart
_gameon:

        ld a, (LevelNumber)
        dec a
        calr InitLevel
        
        ld (Score), 0

        calr DrawStatus
        calr StartTimer

        ld (PlayerPushing), 0
        ld (ControlAnim), 0
        ld (OnToExit), 0
 
_main:

       cp (ControlAnim), 0
       jr nz, Anim

       ; next level?
       cp (OnToExit), 1
       jr nz, ExitCheck1
       calr StopTimer
       cp (LevelNumber), MaxLevels
       jr nz, NextLevelCheck1
       
       ; no more levels - temp display the game over
       calr DoDisplayGameOver
       jp _gamestart
       
NextLevelCheck1:
       inc (LevelNumber)
       jp _levelrestart

ExitCheck1:
       cp (OnToExit), 0
       jr z, Control
       
       ; do player walked into exit animation
       ld (DisableInput), 1
       ld (ExitAnimNo), 2
       calr UpdatePlayerExitAnim
       dec (OnToExit)

Control:
       calr DoControl
       jr _main1

Anim:
       calr DoAnim

_main1:
       inc (ControlAnim)

       ; if we are at the end of the anim cycle then we need to 
       ; do the detections
       cp (ControlAnim), 16
       jr nz, _main1_1
       calr DoGravity
       ld (ControlAnim), 0

_main1_1:
       ld w, 0
       calr vbWait

       cp (ControlAnim), 0
       jr nz, _main2
       cp (CountdownToLevelReset), 0
       jr z, _main2
       dec (CountdownToLevelReset)
       cp (CountdownToLevelReset), 0
       jr nz, _main2
       dec (Lives)
       jp _levelrestart

_main2:

; User shutdown? (Power off pressed?)

        cp      (rUSERS),0
        jp      z,_main      ; no

; Power off NGP

        ld      rw3,VECT_SHUTDOWN
        calr    SYSTEM_CALL

_done:   jr      _done


; -------------------------------
; CONTROL CYCLE - joystick read
; -------------------------------
DoControl:
       ; check timer
       cp (TimeRemaining), 0
       jr nz, DoC_0
       cp (DisableInput), 0 ; only explode once
       jr nz, DoC_0
       
       ; times up - kill player  
       calr DoSuicide

DoC_0:
       cp (Scroll), 1
       jr nz, DoC_1
       
       ; reset scrolling and redraw screem
       calr DrawVisibleLevel

DoC_1:
       calr DoGravity

       ; Update firefly animation
       calr DoFireFlyAnim

       ; animate the exit?
       cp (DiamondsRemaining),0
       jr nz, DoC_11
       calr DoExitAnim

DoC_11:
       ld a, (rSL)
       ld (PlayerMoved), 0
       ld (Scroll), 0

       ; if input is disabled (we are exploding etc.)
       cp (DisableInput), 1
       jr z, DoControlEnd

       cp a, 0
       jr nz, DoC_2
       ld a, (NextMove)
DoC_2:
       ld (PreviousMove), a
       ld (NextMove), 0
       ld (NextMoveOK), 0

       bit PADB_LEFT, a
       jp nz, MoveLeft

       bit PADB_RIGHT, a
       jp nz, MoveRight

       bit PADB_UP, a
       jp nz, MoveUp

       bit PADB_DOWN, a
       jp nz, MoveDown


; move this to title screen only
   IFDEF RUNFROMMENU
      bit PADB_OPTION, a
      jp nz, 200089h   ; return to menu system
   ENDIF
       
       bit PADB_OPTION, a
       jp nz, DoSuicide

       ; if player didn't move make him stand and face forward
       cp (PlayerMoved), 0
       jr nz, DoControlEnd
       ld hl, (PlayerY)
       calr ConvertXYToSpriteXY
       ld a, PLSTAND
       call DrawPlayer

DoControlEnd:
       ret


; -------------------------------
; ANIMATION CYCLE - movement
; -------------------------------
DoAnim:       
       cp (ControlAnim), 8
       jr nz, DoAnimMove
       calr DoGravity

       ; Update firefly animation
       calr DoFireFlyAnim

DoAnimMove:

       ; if player didn't move then don't do second stage of anim
       cp (PlayerMoved), 1
       jr nz, DoAnimMoveEnd

       cp (DisableInput), 1
       jr z, DoAnimMoveEnd
       

       cp (ControlAnim), 4
       jr z, DoAnimMovePlayer
       cp (ControlAnim), 8 
       jr z, DoAnimMovePlayer
       cp (ControlAnim), 12
       jr z, DoAnimMovePlayer
DoAnimEnd:
       cp (Scroll), 1
       jr nz, DoAnimMoveEnd

       ld a, (PreviousMove)

       bit PADB_LEFT, a
       jr z, DAM_1
       dec (rS2SOX)
       jr DAM_4
DAM_1:
       bit PADB_RIGHT, a
       jr z, DAM_2
       inc (rS2SOX)
       jr DAM_4
DAM_2:
       bit PADB_UP, a
       jr z, DAM_3
       dec (rS2SOY)
       jr DAM_4
DAM_3:
       bit PADB_DOWN, a
       jr z, DAM_4
       inc (rS2SOY)
DAM_4:
       
DoAnimMoveEnd:
       ret

DoAnimMovePlayer:
       ld a, (PreviousMove)

       bit PADB_LEFT, a
       jp nz, MoveLeftAnim

       bit PADB_RIGHT, a
       jp nz, MoveRightAnim

       bit PADB_UP, a
       jp nz, MoveUpAnim

       bit PADB_DOWN, a
       jp nz, MoveDownAnim
       
       jp DoAnimEnd


; ---------------------------------------------------------------
; Gravity:
; ROCKS fall down if space under them, fall left or right if space
; left or right and they are balanced on top of other rocks or
; diamonds
; DIAMONDS fall same as rocks but will fall left or right even if
; they are on top of a wall or soil. They require the 3 squares 
; below them to full.
;
; ---------------------------------------------------------------
DoGravity:
   push a
   push xde
   push xhl

   ; load scan position x with 1 and y with total height - 2 (the
   ; bottom line must be solid and the rocks on the line above cannot
   ; fall hense the -2
   ld h, 0
   ld l, (LevelHeight)
   sub l, 2

   ld (InViewY), 0

   ; calculate valid view positions 
   ld a, (VisY)
   ld (TopViewPos), a
   sub (TopViewPos), 2
   jr nc, DoG_Y_1
   ld (TopViewPos), 0
DoG_Y_1:
   ld a, (VisY)
   add a, 11
   ld (BottomViewPos), a
   sub a, l
   jr c, DoG_Y_2
   ld (BottomViewPos), l
DoG_Y_2:

DoGrav_YLoop:
   call ConvertXYtoMap

   ld (InViewX), 0

   ; calculate valid view positions
   ld b, (VisX) ; left pos first (= VisX - 1 or 0)
   sub b, 1
   jr nc, DoG_1
   ld b, 0

DoG_1:
   ld c, (VisX)
   add c, 11

   ; check if the Y range is in view or not
   cp (TopViewPos), l
   jr nz, DoG_Y_3
   ld (InViewY), 0
DoG_Y_3:
   cp (BottomViewPos), l
   jr nz, DoG_Y_4
   ld (InViewY), 1
DoG_Y_4:

   
DoGrav_XLoop:
   ; are we in view?
   cp h, b
   jr nz, DoG_2
   ld (InViewX), 1 ; yes
DoG_2:
   cp h, c ; are we out of view again
   jr nz, DoG_3
   ld (InViewX), 0 ; yes
DoG_3:

   ; check for rock at that position
   ld a, (xde)
   and a, 03Fh ; remove the falling flags

   cp a, TILEEXPLOSION
   jp z, DoExplosion
   cp a, TILEEXPLOSION+1
   jp z, DoExplosion
   cp a, TILEEXPLOSION+2
   jp z, DoExplosion
   cp a, TILEEXPLOSION+3
   jp z, DoExplosion

   ; are we in the control or animation cycle
   cp (ControlAnim), 0
   jr nz, DoGrav_CheckAnim

; -------------------------------
; CONTROL CYCLE
; -------------------------------
   cp a, TILEROCK
   jp z, DoGrav_FoundRock
   cp a, TILEDIAMOND
   jp z, DoGrav_FoundDiamond
   cp a, TILEFIREFLY
   jp z, DoMoveFireFly
   cp a, TILEBUTTERFLY
   jp z, DoMoveButterFly

   jr DoGrav_CheckNext

DoGrav_CheckAnim:
   
   ; are we in the control or animation cycle
   cp (ControlAnim), 8
   jr nz, DoGrav_CheckDetections

; -------------------------------
; ANIMATION CYCLE
; -------------------------------
   ; if there is no falling flag set then skip this tile
   cp a, (xde)
   jr z, DoGrav_CheckAnim1

   cp a, TILEROCK
   jp z, DoGrav_FoundRockFalling
   cp a, TILEDIAMOND
   jp z, DoGrav_FoundDiamondFalling

DoGrav_CheckAnim1:
   cp a, TILEFIREFLYMOVED
   jp z, DoMoveFireFlyAnim
   cp a, TILEBUTTERFLYMOVED
   jp z, DoMoveButterFlyAnim
   
DoGrav_CheckDetections:
   cp (ControlAnim), 16
   jr nz, DoGrav_CheckNext

; -------------------------------
; DETECTION CYCLE
; -------------------------------
   cp a, TILEROCK
   jp z, DoGrav_CheckTileHitSomething
   cp a, TILEDIAMOND
   jp z, DoGrav_CheckTileHitSomething


DoGrav_CheckNext
   inc xde

   inc h
   cp h, (LevelWidth)
   jp nz, DoGrav_XLoop
   
   ; now check previous row
   ld h, 0
   dec l
   cp l, 0
   jp nz, DoGrav_YLoop
   
DoGrav_End:
   pop xhl
   pop xde
   pop a
   ret


; ---------------------------------------------------------------
; CHECK IF FALLING TILE HIT ANYTHING
; ---------------------------------------------------------------
DoGrav_CheckTileHitSomething:

   ; if the player is holding the rock up then don't 
   ; explode him
   cp (xde), a
   jr z, DoGrav_CTHS1

   ; remove the falling flag
   ld (xde), a
   calr CheckPlayerUnderneath

DoGrav_CTHS1:
   calr CheckFireFlyUnderneath

   jp DoGrav_CheckNext




; ---------------------------------------------------------------
; EXPLOSIONS
; ---------------------------------------------------------------
DoExplosion:
   ; change the animation
   inc (xde)

   ; if we have overshot the last animation then clear the explosion
   cp (xde), TILEEXPLOSION+4
   jr nz, DoExp_1
   ld (xde), TILEBLANK

DoExp_1:
   ; --- only do this section if explosion in view ---
   cp (InViewX), 1
   jr nz, DoExp_End
   cp (InViewY), 1
   jr nz, DoExp_End

   ; update screen
   ld a, (xde) ; new value
   call DrawBlock
   
DoExp_End:
   jp DoGrav_CheckNext



; ---------------------------------------------------------------
; ROCK HIT FIREFLY?
; inputs:
;    xde map address of object
; ---------------------------------------------------------------
CheckFireFlyUnderneath:
   push hl
   push xde

   ; make rocks xy match fireflies
   add xde, MAXLEVELWIDTH

   ; did it hit?
   ld a, (xde)
   and a, 15
   cp a, TILEFIREFLY
   jr z, CheckFFU_Hit
   cp a, TILEBUTTERFLY
   jr nz, CheckFFU_End

   ; got the butterfly
   dec h
   calr DrawDiamondExplosion
   jr CheckFFU_End

CheckFFU_Hit: 

   ; got the firefly   
   dec h
   calr DrawExplosion

CheckFFU_End:
   pop xde
   pop hl
   ret



; ---------------------------------------------------------------
; ROCKS FALLING
; ---------------------------------------------------------------
DoGrav_FoundRock:
   push xde

   ; should rock fall down?
   add xde, MAXLEVELWIDTH 
   ld a, (xde)
   cp a, TILEBLANK
   jr z, DoGrav_RockDown
   cp a, SPACERESERVE
   jr z, DoGrav_RockDown

   ; is rock unstable
   cp a, TILEROCK
   jr z, DoGrav_RockLeftOrRight
   cp a, TILEDIAMOND
   jr z, DoGrav_RockLeftOrRight

DoGrav_FoundRockEnd:
   pop xde
   jp DoGrav_CheckNext

DoGrav_RockDown:
   ; hl represents xy of rock

   ; is this space already taken?
   cp a, SPACERESERVE
   jp z, DoGrav_FoundRockEnd


   ; --- only do this section if rock in view ---
   cp (InViewX), 1
   jr nz, FR_1
   cp (InViewY), 1
   jr nz, FR_1

   ; blank rock
   ld a, TILEBLANK
   call DrawBlock
   
   ; draw first animation
   push hl
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEROCK
   call DrawBlockAt
   pop hl
   ; ------------------------------

FR_1:

   ; update map
   call ConvertXYToMap
   ld (xde), TILEROCK
   or (xde), TILEFALLINGDOWN
   add xde, MAXLEVELWIDTH
   ld (xde), SPACERESERVE

DoGrav_RockDownEnd:   
   jp DoGrav_FoundRockEnd

DoGrav_RockLeftOrRight:
   ; hl represent xy of rock
   
   ; check down left
   cp h, 0
   jr z, DoGrav_RLORCheckRight
   
   push hl
   dec h
   inc l
   call ConvertXYToMap
   pop hl

   ld a, (xde)
   cp a, TILEBLANK
   jr z, DoGrav_RLORCheckLeft
   cp a, SPACERESERVE
   jr nz, DoGrav_RLORCheckRight

DoGrav_RLORCheckLeft:

   push xde
   sub xde, MAXLEVELWIDTH
   ld a, (xde)
   pop xde
   cp a, TILEBLANK
   jp z, DoGrav_RockLeft
   cp a, SPACERESERVE
   jr nz, DoGrav_RLORCheckRight

   ; can fall left
   jp DoGrav_RockLeft

DoGrav_RLORCheckRight:
   ; is it too close to right border
   ld a, h
   inc a
   cp a, (LevelWidth)
   jp z, DoGrav_FoundRockEnd
   
   ; check down+right
   push hl
   inc h
   inc l
   call ConvertXYToMap
   pop hl
   
   ld a, (xde)
   cp a, TILEBLANK
   jr z, DoGrav_RLORCheckRightOK
   cp a, SPACERESERVE
   jp nz, DoGrav_FoundRockEnd

DoGrav_RLORCheckRightOK:

   push xde
   sub xde, MAXLEVELWIDTH
   ld a, (xde)
   pop xde
   cp a, TILEBLANK
   jp z, DoGrav_RockRight
   cp a, SPACERESERVE
   jp nz, DoGrav_FoundRockEnd

   ; can fall right
   jp DoGrav_RockRight
   
DoGrav_RockLeft:
   ; hl represents xy of rock

   ; is this square already taken?
   ld a, (xde)
   cp a, SPACERESERVE
   jp z, DoGrav_FoundRockEnd

   ; --- only do this section if rock in view ---
   cp (InViewX), 1
   jr nz, RL_1
   cp (InViewY), 1
   jr nz, RL_1

   ; blank rock
   ld a, TILEBLANK
   call DrawBlock
   
   ; draw first animation
   push hl
   dec h
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEROCK
   call DrawBlockAt
   pop hl
   ; ---------------------------

RL_1:

   ; update map
   call ConvertXYToMap
   or (xde), TILEFALLINGLEFT
   dec xde
   ld (xde), SPACERESERVE
   add xde, MAXLEVELWIDTH
   ld (xde), SPACERESERVE

DoGrav_RockLeftEnd:   
   jp DoGrav_FoundRockEnd


DoGrav_RockRight:
   ; hl represents xy of rock

   ; is this square already taken?
   ld a, (xde)
   cp a, SPACERESERVE
   jp z, DoGrav_FoundRockEnd
   
   ; --- only do this section if rock in view ---
   cp (InViewX), 1
   jr nz, RR_1
   cp (InViewY), 1
   jr nz, RR_1

   ; blank rock
   ld a, TILEBLANK
   call DrawBlock
   
   ; draw first animation
   push hl
   inc h
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEROCK
   call DrawBlockAt
   pop hl
   ; -----------------------------

RR_1:

   ; update map
   call ConvertXYToMap
   or (xde), TILEFALLINGRIGHT
   inc xde
   ld (xde), SPACERESERVE
   add xde, MAXLEVELWIDTH
   ld (xde), SPACERESERVE

DoGrav_RockRightEnd:   
   jp DoGrav_FoundRockEnd


; ** ROCK SECOND ANIMATION **
DoGrav_FoundRockFalling:
   push xde
   
   ld a, (xde)
   and a, TILEFALLINGDOWN
   cp a, TILEFALLINGDOWN
   jr z, DoGrav_RockDownSecondAnim

   ld a, (xde)
   and a, TILEFALLINGLEFT
   cp a, TILEFALLINGLEFT
   jr z, DoGrav_RockLeftSecondAnim
   
   ld a, (xde)
   and a, TILEFALLINGRIGHT
   cp a, TILEFALLINGRIGHT
   jp z, DoGrav_RockRightSecondAnim

DoGrav_FoundRockFallingEnd:
   pop xde
   jp DoGrav_CheckNext


DoGrav_RockDownSecondAnim:
   ; --- only do this section if rock in view ---
   cp (InViewX), 1
   jr nz, RDSecond_1
   cp (InViewY), 1
   jr nz, RDSecond_1

   ; blank current position
   push hl
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEBLANK
   call DrawBlockAt

   ; draw final position
   inc l 
   ld a, TILEROCK
   call DrawBlockAt

   pop hl
   ; --------------------------------

RDSecond_1:
   
   ; erase old position on map
   call ConvertXYToMap
   ld (xde), TILEBLANK
   
   ; update map with new position
   add xde, MAXLEVELWIDTH  
   ld (xde), TILEROCK
   or (xde), TILEFALLINGDOWN

   jp DoGrav_FoundRockFallingEnd


DoGrav_RockLeftSecondAnim:
   ; --- only do this section if rock in view ---
   cp (InViewX), 1
   jr nz, RLSecond_1
   cp (InViewY), 1
   jr nz, RLSecond_1

   ; blank current position
   push hl
   dec h
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEBLANK
   call DrawBlockAt

   ; draw final position
   inc l 
   ld a, TILEROCK
   call DrawBlockAt
   pop hl

   ; ---------------------------
RLSecond_1:  
   
   ; erase old position on map
   call ConvertXYToMap
   ld (xde), TILEBLANK
   dec xde
   ld (xde), TILEBLANK
   
   ; update map with new position
   add xde, MAXLEVELWIDTH  
   ld (xde), TILEROCK
   or (xde), TILEFALLINGDOWN

   jp DoGrav_FoundRockFallingEnd

DoGrav_RockRightSecondAnim:
   ; --- only do this section if rock in view ---
   cp (InViewX), 1
   jr nz, RRSecond_1
   cp (InViewY), 1
   jr nz, RRSecond_1

   ; blank current position
   push hl
   inc h
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEBLANK
   call DrawBlockAt

   ; draw final position
   inc l 
   ld a, TILEROCK
   call DrawBlockAt
   pop hl

   ; ---------------------------
RRSecond_1:  
   
   ; erase old position on map
   call ConvertXYToMap
   ld (xde), TILEBLANK
   inc xde
   ld (xde), TILEBLANK

   ; update map with new position
   add xde, MAXLEVELWIDTH  
   ld (xde), TILEROCK
   or (xde), TILEFALLINGDOWN

   jp DoGrav_FoundRockFallingEnd



; ---------------------------------------------------------------
; DIAMOND FALLING
; ---------------------------------------------------------------
DoGrav_FoundDiamond:
   push xde

   ; should Diamond fall down?
   add xde, MAXLEVELWIDTH 
   ld a, (xde)
   cp a, TILEBLANK
   jr z, DoGrav_DiamondDown
   cp a, SPACERESERVE
   jr z, DoGrav_DiamondDown

   ; diamonds are always unstable
   jr DoGrav_DiamondLeftOrRight

DoGrav_FoundDiamondEnd:
   pop xde
   jp DoGrav_CheckNext

DoGrav_DiamondDown:
   ; hl represents xy of diamond

   ; is this space already taken?
   cp a, SPACERESERVE
   jp z, DoGrav_FoundDiamondEnd

   ; --- only do this section if diamond in view ---
   cp (InViewX), 1
   jr nz, FD_1
   cp (InViewY), 1
   jr nz, FD_1

   ; blank diamond
   ld a, TILEBLANK
   call DrawBlock
   
   ; draw first animation
   push hl
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEDIAMOND
   call DrawBlockAt
   pop hl
   ; ------------------------------

FD_1:

   ; update map
   call ConvertXYToMap
   or (xde), TILEFALLINGDOWN
   add xde, MAXLEVELWIDTH
   ld (xde), SPACERESERVE

DoGrav_DiamondDownEnd:   
   jp DoGrav_FoundDiamondEnd


DoGrav_DiamondLeftOrRight:
   ; hl represent xy of Diamond
   
   ; check down left
   cp h, 0
   jr z, DoGrav_DLORCheckRight
   
   push hl
   dec h
   inc l
   call ConvertXYToMap
   pop hl

   ld a, (xde)
   cp a, TILEBLANK
   jr z, DoGrav_DLORCheckLeft
   cp a, SPACERESERVE
   jr nz, DoGrav_DLORCheckRight

DoGrav_DLORCheckLeft:

   push xde
   sub xde, MAXLEVELWIDTH
   ld a, (xde)
   pop xde
   cp a, TILEBLANK
   jp z, DoGrav_DiamondLeft
   cp a, SPACERESERVE
   jr nz, DoGrav_DLORCheckRight

   ; can fall left
   jp DoGrav_DiamondLeft

DoGrav_DLORCheckRight:
   ; is it too close to right border
   ld a, h
   inc a
   cp a, (LevelWidth)
   jp z, DoGrav_FoundDiamondEnd
   
   ; check down+right
   push hl
   inc h
   inc l
   call ConvertXYToMap
   pop hl
   
   ld a, (xde)
   cp a, TILEBLANK
   jr z, DoGrav_DLORCheckRightOK
   cp a, SPACERESERVE
   jp nz, DoGrav_FoundDiamondEnd

DoGrav_DLORCheckRightOK:

   push xde
   sub xde, MAXLEVELWIDTH
   ld a, (xde)
   pop xde
   cp a, TILEBLANK
   jp z, DoGrav_DiamondRight
   cp a, SPACERESERVE
   jp nz, DoGrav_FoundDiamondEnd

   ; can fall right
   jp DoGrav_DiamondRight
   
DoGrav_DiamondLeft:
   ; hl represents xy of Diamond

   ; is this square already taken?
   ld a, (xde)
   cp a, SPACERESERVE
   jp z, DoGrav_FoundDiamondEnd

   ; blank Diamond
   ld a, TILEBLANK
   call DrawBlock
   
   ; --- only do this section if diamond in view ---
   cp (InViewX), 1
   jr nz, DL_1
   cp (InViewY), 1
   jr nz, DL_1

   ; draw first animation
   push hl
   dec h
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEDIAMOND
   call DrawBlockAt
   pop hl
   ; ---------------------------

DL_1:

   ; update map
   call ConvertXYToMap
   or (xde), TILEFALLINGLEFT
   dec xde
   ld (xde), SPACERESERVE
   add xde, MAXLEVELWIDTH
   ld (xde), SPACERESERVE

DoGrav_DiamondLeftEnd:   
   jp DoGrav_FoundDiamondEnd

DoGrav_DiamondRight:
   ; hl represents xy of Diamond

   ; is this square already taken?
   ld a, (xde)
   cp a, SPACERESERVE
   jp z, DoGrav_FoundDiamondEnd

   ; blank Diamond
   ld a, TILEBLANK
   call DrawBlock
   
   ; --- only do this section if diamond in view ---
   cp (InViewX), 1
   jr nz, DR_1
   cp (InViewY), 1
   jr nz, DR_1

   ; draw first animation
   push hl
   inc h
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEDIAMOND
   call DrawBlockAt
   pop hl
   ; -----------------------------

DR_1:


   ; update map
   
   call ConvertXYToMap
   or (xde), TILEFALLINGRIGHT
   inc xde
   ld (xde), SPACERESERVE
   add xde, MAXLEVELWIDTH
   ld (xde), SPACERESERVE

DoGrav_DiamondRightEnd:   
   jp DoGrav_FoundDiamondEnd

; ** DIAMOND SECOND ANIMATION **
DoGrav_FoundDiamondFalling:
   push xde
   
   ld a, (xde)
   and a, TILEFALLINGDOWN
   cp a, TILEFALLINGDOWN
   jr z, DoGrav_DiamondDownSecondAnim

   ld a, (xde)
   and a, TILEFALLINGLEFT
   cp a, TILEFALLINGLEFT
   jr z, DoGrav_DiamondLeftSecondAnim
   
   ; must be right then
   ld a, (xde)
   and a, TILEFALLINGRIGHT
   cp a, TILEFALLINGRIGHT
   jp z, DoGrav_DiamondRightSecondAnim

DoGrav_FoundDiamondFallingEnd:
   pop xde
   jp DoGrav_CheckNext

DoGrav_DiamondDownSecondAnim:
   ; --- only do this section if diamond in view ---
   cp (InViewX), 1
   jr nz, DDSecond_1
   cp (InViewY), 1
   jr nz, DDSecond_1

   ; blank current position
   push hl
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEBLANK
   call DrawBlockAt

   ; draw final position
   inc l 
   ld a, TILEDIAMOND
   call DrawBlockAt
   pop hl
   ; --------------------------------

DDSecond_1:
   
   ; erase old position on map
   call ConvertXYToMap
   ld (xde), TILEBLANK
   
   ; update map with new position
   add xde, MAXLEVELWIDTH  
   ld (xde), TILEDIAMOND
   or (xde), TILEFALLINGDOWN
   
   jp DoGrav_FoundDiamondFallingEnd

DoGrav_DiamondLeftSecondAnim:
   ; --- only do this section if rock in view ---
   cp (InViewX), 1
   jr nz, DDSecond_1
   cp (InViewY), 1
   jr nz, DDSecond_1

   ; blank current position
   push hl
   dec h
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEBLANK
   call DrawBlockAt

   ; draw final position
   inc l 
   ld a, TILEDIAMOND
   call DrawBlockAt
   pop hl
   
   ; ---------------------------
DLSecond_1:  

   ; erase old position on map
   call ConvertXYToMap
   ld (xde), TILEBLANK
   dec xde
   ld (xde), TILEBLANK ; remove reserved space to left
   
   ; update map with new position
   add xde, MAXLEVELWIDTH  
   ld (xde), TILEDIAMOND
   or (xde), TILEFALLINGDOWN
   
   jp DoGrav_FoundDiamondFallingEnd

DoGrav_DiamondRightSecondAnim:
   ; --- only do this section if diamond in view ---
   cp (InViewX), 1
   jr nz, DRSecond_1
   cp (InViewY), 1
   jr nz, DRSecond_1

   ; blank current position
   push hl
   inc h
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc l
   ld a, TILEBLANK
   call DrawBlockAt

   ; draw final position
   inc l 
   ld a, TILEDIAMOND
   call DrawBlockAt
   pop hl
   ; ---------------------------
DRSecond_1:  
   
   ; erase old position on map
   call ConvertXYToMap
   ld (xde), TILEBLANK
   inc xde
   ld (xde), TILEBLANK ; remove reserved space to left
   
   ; update map with new position
   add xde, MAXLEVELWIDTH  
   ld (xde), TILEDIAMOND
   or (xde), TILEFALLINGDOWN
   
   jp DoGrav_FoundDiamondFallingEnd


; ---------------------------------------------------------------
; User pressed option - do suicide
; ---------------------------------------------------------------
DoSuicide:
   push xde
   push hl

   ldw hl, (PlayerY)
   add h, (VisX)
   add l, (VisY)

   ; remove player sprite from screen
   push h
   ld h, 200
   calr DrawPlayer
   pop h

   ; draw explosion
   dec l
   dec h
   calr DrawExplosion

   ; set we're exploding flag to disable further joystick
   ld (DisableInput), 1
   ld (CountdownToLevelReset), 8

   pop hl
   pop xde
   ret


; ---------------------------------------------------------------
; Check Player Underneath - is the player underneath the falling
; object?
; inputs:
;    xde map address of object
; ---------------------------------------------------------------
CheckPlayerUnderneath:
   push hl
   push xde

   ; make rocks xy match players xy
   add xde, MAXLEVELWIDTH

   ; did it hit?
   cp (xde), PLSTAND
   jr nz, CheckPU_End

   ; remove player sprite from screen
   ldw hl, (PlayerY)
   push h
   ld h, 200
   calr DrawPlayer
   pop h

   ; ensure the player no longer exist
   ld (xde), TILEBLANK

   ; draw explosion
   add h, (VisX)
   add l, (VisY)
   dec l
   dec h
   calr DrawExplosion

   ; set we're exploding flag to disable further joystick
   ld (DisableInput), 1
   ld (CountdownToLevelReset), 8
   

CheckPU_End
   pop xde
   pop hl
   ret

; ---------------------------------------------------------------
; DrawExplosion - takes hl as xy of top left of explosion and
; draw a 3x3 explosion. Also updates map.
; If the tile where the explosion would be is solid wall then
; that tile is missed off.
; ---------------------------------------------------------------
DrawExplosion:
   push hl
   push bc
   push xde

   ld c, 0
DrawExp_YLoop:   
   ld b, 0

DrawExp_XLoop:
   calr ConvertXYtoMap
   
   cp (xde), TILESOLIDWALL ; don't explode solid wall
   jr z, DrawExp_NextX
   cp (xde), TILEEXIT
   jr z, DrawExp_NextX

   ld a, TILEEXPLOSION
   calr DrawBlock

   ; was player hit?
   cp (xde), PLSTAND
   jr nz, DrawExp_1
   ld (xde), a
   push hl
   push h
   ld h, 200
   calr DrawPlayer ; remove player from screen
   pop h

   dec h
   dec l
   calr DrawExplosion
   ; set we're exploding flag to disable further joystick
   ld (DisableInput), 1
   ld (CountdownToLevelReset), 8
   pop hl

DrawExp_1:
   ; was firefly underneath
   ld w, (xde)
   and w, 15
   cp w, TILEFIREFLY
   jr nz, DrawExp_2

   ld (xde), a
   push hl
   dec h
   dec l
   calr DrawExplosion
   pop hl

DrawExp_2:
   ; was butterfly underneath
   ld w, (xde)
   and w, 15
   cp w, TILEBUTTERFLY
   jr nz, DrawExp_3

   ld (xde), a
   push hl
   dec h
   dec l
   calr DrawDiamondExplosion
   pop hl

DrawExp_3:
   ld (xde), a

DrawExp_NextX:
   inc h
   inc b
   cp b, 3
   jr nz, DrawExp_XLoop

   sub h, 3
   inc l
   inc c
   cp c, 3
   jr nz, DrawExp_YLoop

   pop xde
   pop bc
   pop hl
   ret





; ---------------------------------------------------------------
; Joystick Handling
;
; ---------------------------------------------------------------

MoveRight:
   push xwa
   push xde
   push xhl

   ldw hl, (PlayerY)
   inc h
   add h, (VisX)
   add l, (VisY)

   ; check if move posible
   call ConvertXYtoMap
   ld a, (xde)
   cp a, TILESOIL 
   jr z, MoveRightOK
   cp a, TILEBLANK
   jr z, MoveRightOK
   cp a, TILEDIAMOND
   jr z, MoveRightOK
   cp a, TILEROCK
   jr z, MoveRightPushRock
   cp a, TILEEXIT
   jr nz, MoveRightExitEnd
   cp (DiamondsRemaining), 0
   jr nz, MoveRightExitEnd
   ld (OnToExit), 4
   jr MoveRightOK
   
MoveRightExitEnd:
   jr MoveRightEnd

MoveRightOK:
   ; check if we have just collected a diamond
   cp a, TILEDIAMOND
   call z, CollectDiamond

   ; blank new position if not already blank
   cp a, TILEBLANK
   jr z, MROK1
   cp a, TILEEXIT
   jr z, MROK1
   ld a, TILEBLANK
   calr DrawBlock
   ld (xde), TILEBLANK
MROK1:

   ld a, (rSL)
   bit PADB_A, a
   jr nz, MoveRightEnd

   ; move or scroll
   ld a, (ScrollSizeX)
   cp (VisX), a
   jr z, MRO_1
   cp (PlayerX), 5
   jr nz, MRO_1
   ld (Scroll), 1

MRO_1:
   ; draw the first animation
   ldw hl, (PlayerY)
   call ConvertXYToSpriteXY

   cp (Scroll), 1
   jr z, MRO_2
   add h, 4
MRO_2:
   ldw (AnimY), hl
   ld a, PLRIGHT
   call DrawPlayer

   ; store player on map
   dec xde
   ld (xde), SPACERESERVE
   inc xde
   ld (xde), PLSTAND

   ld (PlayerMoved), 1

MoveRightEnd:
   pop xhl
   pop xde
   pop xwa
   jp DoControlEnd


MoveRightPushRock:
   ; first check is push posible
   calr PushRockRightOK
   jr nc, MoveRightEnd

   ; draw the first animation
   ld a, TILEBLANK
   calr DrawBlock
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   inc h
   ld a, TILEROCK
   calr DrawBlockAt

   ; move or scroll
   ld a, (ScrollSizeX)
   cp (VisX), a
   jr z, MRPR_1
   cp (PlayerX), 5
   jr nz, MRPR_1
   ld (Scroll), 1

MRPR_1:
   ldw hl, (PlayerY)
   call ConvertXYToSpriteXY

   cp (Scroll), 1
   jr z, MRPR_2
   add h, 4
MRPR_2:

   add h, 4
   ldw (AnimY), hl
   ld a, PLRIGHT
   call DrawPlayer
   
   
   ; store player on map and the fact that the rock is moving
   dec xde
   ld (xde), SPACERESERVE
   inc xde
   ld (xde), TILEROCKPUSHED
   inc xde
   ld (xde), SPACERESERVE
   ld (PlayerMoved), 1
   ld (PlayerPushing), 1
   
   jr MoveRightEnd


PushRockRightOK:
   push a
   push xde
   push hl

   ; if fire pressed then fail
   ld a, (rSL)
   bit PADB_A, a
   jr nz, PushRightFail

   ; too close to edge?
   inc h
   cp h, (LevelWidth)
   jr z, PushRightFail
   
   ; ensure other side of rock is blank
   inc xde
   cp (xde), TILEBLANK
   jr nz, PushRightFail

   ; push right OK
   scf
   jr PushRightTidyUp

PushRightFail:
   rcf

PushRightTidyUp:
   pop hl
   pop xde
   pop a
   ret
   

MoveRightAnim:
   push xwa
   push xde
   push xhl

   cp (Scroll), 1
   jr z, MRA_1

   add (AnimX), 4
MRA_1:
   ldw hl, (AnimY)

   ; draw the new animation
   ld a, PLRIGHT
   call DrawPlayer

   ; erase the old position on the map
   cp (ControlAnim), 8
   jr nz, MoveRightAnimEnd

   ldw hl, (PlayerY)
   add h, (VisX)
   add l, (VisY)
   calr ConvertXYtoMap
   ; old position may have already gone (explosion etc)
   cp (xde), SPACERESERVE
   jr nz, MRA_11
   ld (xde), TILEBLANK
MRA_11:

   cp (PlayerPushing), 1
   jr nz, MRA_2

   ; player was pushing a rock
   inc h
   ld a, TILEBLANK
   calr DrawBlock
   inc h
   ld a, TILEROCK
   calr DrawBlock

   ; now update map
   calr ConvertXYtoMap
   ld (xde), TILEROCK
   dec xde
   ld (xde), PLSTAND

   ld (PlayerPushing), 0

MRA_2:
   cp (Scroll), 1
   jr z, MRA_3
   inc (PlayerX)
   jr MoveRightAnimEnd
MRA_3
   inc (VisX)

MoveRightAnimEnd:
   pop xhl
   pop xde
   pop xwa
   jp DoAnimEnd
   

MoveLeft:
   push xwa
   push xde
   push xhl

   ldw hl, (PlayerY)
   dec h
   add h, (VisX)
   add l, (VisY)

   ; check if move posible
   calr ConvertXYtoMap
   ld a, (xde)
   cp a, TILESOIL 
   jr z, MoveLeftOK
   cp a, TILEBLANK
   jr z, MoveLeftOK
   cp a, TILEDIAMOND
   jr z, MoveLeftOK
   cp a, TILEROCK
   jr z, MoveLeftPushRock
   cp a, TILEEXIT
   jr nz, MoveLeftExitEnd
   cp (DiamondsRemaining), 0
   jr nz, MoveLeftExitEnd
   ld (OnToExit), 4
   jr MoveLeftOK

MoveLeftExitEnd:
   jr MoveLeftEnd

MoveLeftOK:
   ; check if we have just collected a diamond
   cp a, TILEDIAMOND
   call z, CollectDiamond

   ; blank new position if it isn't already blank
   cp a, TILEBLANK
   jr z, MLOK1
   cp a, TILEEXIT
   jr z, MLOK1
   ld a, TILEBLANK
   calr DrawBlock
   ld (xde), TILEBLANK
MLOK1:

   ld a, (rSL)
   bit PADB_A, a
   jr nz, MoveLeftEnd

   ; move or scroll
   cp (VisX), 0
   jr z, MLO_1
   cp (PlayerX), 5
   jr nz, MLO_1
   ld (Scroll), 1

MLO_1:
   ; draw first anim
   ldw hl, (PlayerY)
   call ConvertXYToSpriteXY

   cp (Scroll), 1
   jr z, MLO_2
   sub h, 4
MLO_2:
   ldw (AnimY), hl
   ld a, PLLEFT
   call DrawPlayer

   ; store player on map
   inc xde
   ld (xde), SPACERESERVE
   dec xde
   ld (xde), PLSTAND

   ld (PlayerMoved), 1

MoveLeftEnd:
   pop xhl
   pop xde
   pop xwa
   jp DoControlEnd

MoveLeftPushRock:
   ; first check is push posible
   calr PushRockLeftOK
   jr nc, MoveLeftEnd

   ; draw the first animation
   ld a, TILEBLANK
   calr DrawBlock
   sub h, (VisX)
   sub l, (VisY)  
   sll 1, h
   sll 1, l
   dec h
   ld a, TILEROCK
   calr DrawBlockAt

   ; move or scroll
   cp (VisX), 0
   jr z, MLPR_1
   cp (PlayerX), 5
   jr nz, MLPR_1
   ld (Scroll), 1

MLPR_1:
   ldw hl, (PlayerY)
   call ConvertXYToSpriteXY

   cp (Scroll), 1
   jr z, MLPR_2
   sub h, 4
MLPR_2:
   ldw (AnimY), hl
   ld a, PLLEFT
   call DrawPlayer
   
   ; store player on map and the fact that the rock is moving
   inc xde
   ld (xde), SPACERESERVE
   dec xde
   ld (xde), TILEROCKPUSHED
   dec xde
   ld (xde), SPACERESERVE
   ld (PlayerMoved), 1
   ld (PlayerPushing), 1
   
   jr MoveLeftEnd


PushRockLeftOK:
   push a
   push xde
   push hl

   ; if fire pressed then fail
   ld a, (rSL)
   bit PADB_A, a
   jr nz, PushLeftFail

   ; too close to edge?
   dec h
   cp h, 0
   jr z, PushLeftFail
   
   ; ensure other side of rock is blank
   dec xde
   cp (xde), TILEBLANK
   jr nz, PushLeftFail

   ; push left OK
   scf
   jr PushleftTidyUp

PushleftFail:
   rcf

PushleftTidyUp:
   pop hl
   pop xde
   pop a
   ret

MoveLeftAnim:
   push xwa
   push xde
   push xhl

   cp (Scroll), 1
   jr z, MLA_1

   sub (AnimX), 4
MLA_1:
   ldw hl, (AnimY)

   ; draw the new animation
   ld a, PLLEFT
   call DrawPlayer

   ; erase the old position on the map
   cp (ControlAnim), 8
   jr nz, MoveLeftAnimEnd

   ; erase the old position on the map
   ldw hl, (PlayerY)
   add h, (VisX)
   add l, (VisY)
   calr ConvertXYtoMap

   ; old position may have already gone (explosion etc)
   cp (xde), SPACERESERVE
   jr nz, MLA_11
   ld (xde), TILEBLANK
MLA_11:

   cp (PlayerPushing), 1
   jr nz, MLA_2

   ; player was pushing a rock
   dec h
   ld a, TILEBLANK
   calr DrawBlock
   dec h
   ld a, TILEROCK
   calr DrawBlock

   ; now update map
   calr ConvertXYToMap
   ld (xde), TILEROCK
   inc xde
   ld (xde), PLLEFT

   ld (PlayerPushing), 0

MLA_2
   cp (Scroll), 1
   jr z, MLA_3
   dec (PlayerX)
   jr MoveLeftAnimEnd
MLA_3:
   dec (VisX)

MoveLeftAnimEnd:
   pop xhl
   pop xde
   pop xwa
   jp DoAnimEnd



MoveDown:
   push xwa
   push xde
   push xhl

   ldw hl, (PlayerY)
   inc l
   add h, (VisX)
   add l, (VisY)
 
   ; check if move posible
   calr ConvertXYtoMap
   ld a, (xde)
   cp a, TILESOIL 
   jr z, MoveDownOK
   cp a, TILEBLANK
   jr z, MoveDownOK
   cp a, TILEDIAMOND
   jr z, MoveDownOK
   cp a, TILEEXIT
   jr nz, MoveDownExitEnd
   cp (DiamondsRemaining), 0
   jr nz, MoveDownExitEnd
   ld (OnToExit), 4
   jr MoveDownOK
   
MoveDownExitEnd:
   jr MoveDownEnd

MoveDownOK:
   ; check if we have just collected a diamond
   cp a, TILEDIAMOND
   call z, CollectDiamond

   ; blank new position if not already blank
   cp a, TILEBLANK
   jr z, MDOK1
   cp a, TILEEXIT
   jr z, MDOK1
   ld a, TILEBLANK
   calr DrawBlock
   ld (xde), TILEBLANK
MDOK1:

   ld a, (rSL)
   bit PADB_A, a
   jr nz, MoveDownEnd

   ; move or scroll
   ld a, (ScrollSizeY)
   cp (VisY), a
   jr z, MDO_1
   cp (PlayerY), 5
   jr nz, MDO_1
   ld (Scroll), 1

MDO_1:
   ; draw first anim
   ldw hl, (PlayerY)
   call ConvertXYToSpriteXY

   cp (Scroll), 1
   jr z, MDO_2
   add l, 4
MDO_2:
   ldw (AnimY), hl
   ld a, PLDOWN
   call DrawPlayer

   ; store player on map and remove previous player
   sub xde, MAXLEVELWIDTH
   ld (xde), SPACERESERVE
   add xde, MAXLEVELWIDTH
   ld (xde), PLSTAND

   ld (PlayerMoved), 1

MoveDownEnd:
   pop xhl
   pop xde
   pop xwa
   jp DoControlEnd


MoveDownAnim:
   push xwa
   push xde
   push xhl
   
   cp (Scroll), 1
   jr z, MDA_1

   add (AnimY), 4
MDA_1:
   ldw hl, (AnimY)

   ; draw the new animation
   ld a, PLDOWN
   call DrawPlayer

   ; erase the old position on the map
   cp (ControlAnim), 8
   jr nz, MoveDownAnimEnd

   ; erase the old position on the map
   ldw hl, (PlayerY)
   add h, (VisX)
   add l, (VisY)
   calr ConvertXYtoMap

   ; old position may have already gone (explosion etc)
   cp (xde), SPACERESERVE
   jr nz, MDA_11
   ld (xde), TILEBLANK
MDA_11:
   
   cp (Scroll), 1
   jr z, MDA_2
   inc (PlayerY)
   jr MoveDownAnimEnd
MDA_2:
   inc (VisY)

MoveDownAnimEnd:
   pop xhl
   pop xde
   pop xwa
   jp DoAnimEnd
   

MoveUp:
   push xwa
   push xde
   push xhl

   ldw hl, (PlayerY)
   dec l
   add h, (VisX)
   add l, (VisY)

   ; check if move posible
   calr ConvertXYtoMap
   ld a, (xde)
   cp a, TILESOIL 
   jr z, MoveUpOK
   cp a, TILEBLANK
   jr z, MoveUpOK
   cp a, TILEDIAMOND
   jr z, MoveUpOK
   cp a, TILEEXIT
   jr nz, MoveUpExitEnd
   cp (DiamondsRemaining), 0
   jr nz, MoveUpExitEnd
   ld (OnToExit), 4
   jr MoveUpOK
   
MoveUpExitEnd:
   jr MoveUpEnd

MoveUpOK:
   ; check if we have just collected a diamond
   cp a, TILEDIAMOND
   call z, CollectDiamond

   ; blank new position if not already blank
   cp a, TILEBLANK
   jr z, MUOK1
   cp a, TILEEXIT
   jr z, MUOK1
   ld a, TILEBLANK
   calr DrawBlock
   ld (xde), TILEBLANK

MUOK1:
   ld a, (rSL)
   bit PADB_A, a
   jr nz, MoveUpEnd

   ; move or scroll
   cp (VisY), 0
   jr z, MUO_1
   cp (PlayerY), 5
   jr nz, MUO_1
   ld (Scroll), 1

MUO_1:
   ; draw first anim
   ldw hl, (PlayerY)
   call ConvertXYToSpriteXY

   cp (Scroll), 1
   jr z, MUO_2
   sub l, 4
MUO_2:
   ldw (AnimY), hl
   ld a, PLUP
   call DrawPlayer

   ; store player on map
   add xde, MAXLEVELWIDTH
   ld (xde), SPACERESERVE
   sub xde, MAXLEVELWIDTH
   ld (xde), PLSTAND

   ld (PlayerMoved), 1

MoveUpEnd:
   pop xhl
   pop xde
   pop xwa
   jp DoControlEnd

MoveUpAnim:
   push xwa
   push xde
   push xhl

   cp (Scroll), 1
   jr z, MUA_1

   sub (AnimY), 4
MUA_1:
   ldw hl, (AnimY)

   ; draw the new animation
   ld a, PLUP
   call DrawPlayer

   ; erase the old position on the map
   cp (ControlAnim), 8
   jr nz, MoveUpAnimEnd

   ; erase the old position on the map
   ldw hl, (PlayerY)
   add h, (VisX)
   add l, (VisY)
   calr ConvertXYtoMap

   ; old position may have already gone (explosion etc)
   cp (xde), SPACERESERVE
   jr nz, MUA_11
   ld (xde), TILEBLANK
MUA_11:
   
   cp (Scroll), 1
   jr z, MUA_2
   dec (PlayerY)
   jr MoveUpAnimEnd
MUA_2:
   dec (VisY)

MoveUpAnimEnd:
   pop xhl
   pop xde
   pop xwa
   jp DoAnimEnd


; ---------------------------------------------------------------
; When a diamond has been collected the diamond remaining count
; is reduced until zero is reached. At this point the exit opens
; Score is also increased.
; ---------------------------------------------------------------
CollectDiamond:
   push xde
   cp (DiamondsRemaining), 0
   jr z, CD_1
   dec (DiamondsRemaining)
   calr DrawDiamondsRemaining

   ; check if the exit need opening
   ; ...

CD_1:
   ; update score
   ld xde, (Score)
   add xde, 10
   ld (Score), xde
   calr DrawScore
   pop xde
   ret



; ---------------------------------------------------------------
; Initialise Memory before game starts
;
; ---------------------------------------------------------------
InitialiseMemory:
; Clear Sprite Attribute Memory

        ld      wa,0
        ld      b,64
        ld      xix,_SPRRAM
SAMloop:
        ld      (xix+),wa
        ld      (xix+),wa
        djnz    b,SAMloop

; Clear Scroll 1 & 2 VRAM with 0
	calr ClearScreen
; clear RAM
        ld      xde, _MAINRAM
InitMem1:
        ld      (xde), 0
        inc     xde
        cp      xde, _MAINRAM + 800h
        jr      nz, InitMem1

        ret

; ---------------------------------------------------------------
; Clears the screen
;----------------------------------------------------------------
ClearScreen:
        ld      xhl,_SCR1RAM
        ld      wa, 0
        ld      (xhl),wa
        ld      xde,xhl
        inc     xde
        inc     xde
        ld      bc,800h-1
        ldirw
	ret


; ---------------------------------------------------------------
; DrawPlayer -draws a 2x2 player using tile coordinates
; hl = xy, a=amimation
; 
; ---------------------------------------------------------------
DrawPlayer:
   push xwa
   push xbc
   push xhl

   ; first copy animation to player
   calr CopyAnim2Player

   ; draw the sprite
   ldw wa, TILEPLAYER
   ld c, PLAYERSPRNO
   ld b, 0
   calr DrawSprite
   inc c
   add h, 8
   inc wa
   calr DrawSprite
   inc c
   sub h, 8
   add l, 8
   inc wa
   calr DrawSprite
   inc c
   add h, 8
   inc wa
   calr DrawSprite
   pop xhl
   pop xbc
   pop xwa
   ret

CopyAnim2Player:
   push a
   push bc
   push xde
   push xhl

   ld xhl, 0
   ld l, (ControlAnim)
   srl 2, l
   add l, a
   sub l, PLAYERBASE
   sll 6, xhl
   add xhl, PLAYERANIM
   ld bc, 32 ; 4 chars
   ld xde, _TILERAM + (TILEPLAYER * 16)
   ldirw (xde+),(xhl+)
       
   pop xhl
   pop xde
   pop bc
   pop a
   ret

DoExitAnim
   push a
   push bc
   push xde
   push xhl

   ld xhl, 0
   ld l, (ExitAnimNo)
   sll 6, xhl
   add xhl, EXITANIM
   ld bc, 32 ; 4 chars
   ld xde, _TILERAM + (TILEEXIT * 64)
   ldirw (xde+),(xhl+)

   inc (ExitAnimNo)
   and (ExitAnimNo), 3

DoExit_End:
   pop xhl
   pop xde
   pop bc
   pop a
   ret
   

UpdatePlayerExitAnim:
   push hl
   ldw hl, (PlayerY)
   call ConvertXYToSpriteXY

   cp (OnToExit), 4
   jr nz, UPEA_3
   ld a, PLUP+2
   jr UPEA_End

UPEA_3:
   cp (OnToExit), 3
   jr nz, UPEA_2
   ld a, PLEXIT
   jr UPEA_End
   
UPEA_2:
   cp (OnToExit), 2
   jr nz, UPEA_End1
   ld a, PLEXIT + 1

UPEA_End:
   calr DrawPlayer
UPEA_End1:
   pop hl
   ret


DoFireFlyAnim:
   push a
   push bc
   push xde
   push xhl

   ; fire fly
   ld xhl, 0
   ld l, (FireFlyAnimNo)
   sll 6, xhl
   add xhl, FIREFLYANIM
   ld bc, 32 ; 4 chars
   ld xde, _TILERAM + (TILEFIREFLY * 64)
   ldirw (xde+),(xhl+)

   ; butter fly
   ld xhl, 0
   ld l, (ButterFlyAnimNo)
   sll 6, xhl
   add xhl, BUTTERFLYANIM
   ld bc, 32 ; 4 chars
   ld xde, _TILERAM + (TILEBUTTERFLY * 64)
   ldirw (xde+),(xhl+)

   ; inc butter fly anim number 0-3
   inc (ButterFlyAnimNo)
   and (ButterFlyAnimNo), 3

   ; inc the firefly anim number 0-5
   inc (FireFlyAnimNo)
   cp (FireFlyAnimNo), 6
   jr nz, DoFireFly_End
   ld (FireFlyAnimNo), 0

DoFireFly_End:
   pop xhl
   pop xde
   pop bc
   pop a
   ret




; ---------------------------------------------------------------
; 
; DrawSprite
;
; Draws a Sprite
; inputs: wa = tile, hl = xy
; c = sprite #, b = palette #
;
; ---------------------------------------------------------------
DrawSprite
    push xde
    push xhl
    push xwa

    ; sprite registers for specified sprite
    ld xde, 0
    ld e, c
    sll 2, xde 
    add xde, _SPRRAM

    add w, 24 ; set priority
    ldw (xde), wa
    add xde,2
    ld (xde), h
    inc xde
    ld (xde), l

    ld xde, 0
    ld e, c
    add xde, 8c00h
    ld (xde), b

    pop xwa
    pop xhl
    pop xde
    ret


; ---------------------------------------------------------------
; DrawBlock -draws a 2x2 block using tile coordinates
; hl = xy, a=block no
; 
; ---------------------------------------------------------------
DrawBlock:
   push hl
   push wa
   push xde

   ; convert tile positions to screen positions
   sub h, (VisX)
   sll 1, h
   add h, 2
   sub l, (VisY)
   sll 1, l
   add l, 2
   
   ; get tile colour
   ld xde, 0
   ld e, a
   add xde, TILECOLOURS
   ld w, (xde)
   sll 1, w

   sll 2, a
   calr PrintChar
   inc h
   inc a
   calr PrintChar
   dec h
   inc l
   inc a
   calr PrintChar
   inc h 
   inc a
   calr PrintChar

   pop xde
   pop wa
   pop hl
   ret



; ---------------------------------------------------------------
; DrawBlockAt -draws a 2x2 block using screen coordinates
; hl = xy, a=block no
; 
; ---------------------------------------------------------------
DrawBlockAt:
   push hl
   push wa
   push xde

   add h, 2
   add l, 2

   ; get tile colour
   ld xde, 0
   ld e, a
   add xde, TILECOLOURS
   ld w, (xde)
   sll 1, w

   sll 2, a
   calr PrintChar
   inc h
   inc a
   calr PrintChar
   dec h
   inc l
   inc a
   calr PrintChar
   inc h 
   inc a
   calr PrintChar

   pop xde
   pop wa
   pop hl
   ret


DrawBlockRaw:
   push hl
   push wa
   push xbc
   push xde

   ; convert tile positions to screen positions
   ld xbc, 0
   ld c, l
   sll 7, xbc
   add xbc, _SCR2RAM
   ld xde, 0
   sll 2, h
   ld e, h
   add xbc, xde
   
   ; get tile colour
   ld xde, 0
   ld e, a
   and e, 0fh
   add xde, TILECOLOURS
   ld w, (xde)
   sll 1, w
   sll 2, a

   ldw (xbc+), wa
   inc a
   ldw (xbc), wa
   add xbc, 62
   inc a
   ldw (xbc+), wa
   inc a
   ldw (xbc), wa

   pop xde
   pop xbc
   pop wa
   pop hl
   ret


; ---------------------------------------------------------------
; PrintChar
; hl = xy, wa=char
; 
; ---------------------------------------------------------------
PrintChar:
    push xhl
    push xde
    push xbc
    ld xde, _SCR2RAM
    ld c, h
    ld b, 0
    sll 1, xbc
    add de, bc
    ld h, 0
    sll 6, xhl
    add de, hl
    ldw (xde), wa
    pop xbc
    pop xde
    pop xhl
    ret


; ---------------------------------------------------------------
; PrintCharTop - like PrintChar but uses top plane.
; hl = xy, wa=char
; 
; ---------------------------------------------------------------
PrintCharTop:
    push xhl
    push xde
    push xbc
    ld xde, _SCR1RAM
    ld c, h
    ld b, 0
    sll 1, xbc
    add de, bc
    ld h, 0
    mul xhl, 40h
    add de, hl
    ld (xde), a
    add xde, 1
    ld (xde), w
    pop xbc
    pop xde
    pop xhl
    ret


; ---------------------------------------------------------------
; InitLevel - copies the level to ram, initialises the player
; x and y cordinates and draws the screen
;
; inputs a = level number
; ---------------------------------------------------------------
InitLevel:
   push xwa
   push xbc
   push xde
   push xhl
   push xix

   ; find level data
   ld xde, 0
   ld e, a
   sll 2, e
   add xde, TheLevels
   ld xbc, (xde) ; xbc now points to level

   ; get size
   ld d, (xbc+)
   ld (LevelWidth), d
   sub d, 10 ; get the scrolling size (level witdh - screensize (10))
   ld (ScrollSizeX), d
   ld d, (xbc+)
   ld (LevelHeight), d
   sub d, 10 ; get the scrolling size (level height - screensize (10))
   ld (ScrollSizeY), d

   ; get player x and y
   ld d, (xbc+)
   ld (PlayerX), d
   ld d, (xbc+)
   ld (PlayerY), d

   ; get the number of diamonds
   ld d, (xbc+)
   ld (DiamondsRemaining), d

   ; get the time limit
   ld d, (xbc+)
   ld (TimeRemaining), d

   ld xde, THISLEVEL

   ; generate padding size
   ld xhl, 0
   ld l, MAXLEVELWIDTH
   sub l, (LevelWidth)
   ld xix, xhl

   ; initialise loop x and y
   ld xhl, 0

InitLevel_1:
   ld a, (xbc+)
   sub a, '0'
   ld (xde+), a
   inc h
   cp h, (LevelWidth)
   jr nz, InitLevel_1
   
   add xde, xix ; add padding
   ld h, 0
   inc l
   cp l, (LevelHeight)
   jr nz, InitLevel_1

   ld (LevelName), xbc

   ; sort the players X position out
   ld w, (PLAYERX)
   ld a, (LEVELWIDTH)
   calr CalcPos
   ld (PLAYERX), b
   ld (VisX), c
   
   ; sort the players Y position out
   ld w, (PLAYERY)
   ld a, (LEVELHEIGHT)
   calr CalcPos
   ld (PLAYERY), b
   ld (VisY), c

   calr DisplayLevelName

   calr DrawVisibleLevel

   ; draw player
   ldw hl, (PlayerY)
   calr ConvertXYToSpriteXY
   ld a, PLSTAND
   call DrawPlayer

   ; store player
   ldw hl, (PlayerY)
   add h, (VisX)
   add l, (VisY)
   calr ConvertXYtoMap
   ld (xde), PLSTAND

   pop xix
   pop xhl
   pop xde
   pop xbc
   pop xwa
   ret

; ---------------------------------------------------------------
; DisplayLevelName
; Coarse scrolls the level name in from the right, waits a while,
; then scrolls is off to the left before returning
; ---------------------------------------------------------------
DisplayLevelName:
   push xwa
   push xbc
   push xde
   push xhl

;first clear the screen
   calr ClearScreen

; reset scrolling area
   ld (rS2SOX), 0
   ld (rS2SOY), 0

; shift player off screen
   ldw hl, 0a0a0h
   calr DrawPlayer


; draw the word "level"
   ld b, 1
   ldw hl, 0607h
   ld xde, strLevel
   calr RenderString

; add the level number
   ld xbc, 0
   ld c, (LevelNumber)
   div xbc, 10
   ld de, qbc
   add e, '0'
   add c, '0'
   ld (tmpLevelNumberStr), c
   ld (tmpLevelNumberStr + 1), e
   ld (tmpLevelNumberStr + 2), 0

   ld b, 1
   ldw hl, 0c07h
   ld xde, tmpLevelNumberStr
   calr RenderString

; now add the level name

   ld b, 3
   ldw hl, 000ah
   ld xde, (LevelName)
   calr RenderString

   ld w, 2
   calr Sleep

   pop xhl
   pop xde
   pop xbc
   pop xwa

   ret

strLevel db "LEVEL",0



; ---------------------------------------------------------------
; DrawVisibleLevel
; Draws the level on the screen. The screen should be scrolled to
; 16 pixels in (both x and y). For the NGPCs 10x9.5 visible area
; this draws 12x12 so that the screen can be scrolled in any 
; direction by 16 pixels (one tile) before having to be redrawn
;
; VisX and VisY is used for the offset into the level to draw from
; If these values are 0 then nothing is drawn on the first row and
; column of the 12x12 on the ground that there is no level there
; to scroll to.
; ---------------------------------------------------------------
DrawVisibleLevel:
   push xwa
   push xbc
   push xde
   push xhl
   push xix

   ; reset all of the register
   ld a, 16
   ld (rS2SOX), a
   ld (rS2SOY), a

   ; generate padding size
   ld xix, 20

   ld d, (VisX)
   sub d, 1
   jr nc, DVL_GP_1
   inc xix
DVL_GP_1:
   ld e, (VisY)
   dec e

; set xbc to point to start of level data
; if x or y = -1 then ignore 
   ld xbc, 0
   cp e, -1
   jr z, DVL_1
   ld c, e
   sll 5, xbc
DVL_1:
   cp d, -1
   jr z, DVL_2
   ld xwa,0
   ld a, d
   add xbc, xwa
DVL_2:
   add xbc, THISLEVEL

   ldw hl, 0 ; actual screen pos.

DVL_ROW:
   cp e, -1
   jr z, DVL_4

DVL_COLUMN:
   cp d, -1
   jr z, DVL_3
   ld a, (xbc+)
   cp a, PLSTAND
   jr nz, DVL_5
   ld a, TILEBLANK
DVL_5:
   calr DrawBlockRaw
DVL_3:
   inc d
   inc h
   cp h, 12
   jr nz, DVL_COLUMN

   add xbc, xix ; add padding

DVL_4:
   inc e
   ld d, (VisX)
   dec d
   ld h, 0
   inc l
   cp l , 12
   jr nz, DVL_ROW

   pop xix
   pop xhl
   pop xde
   pop xbc
   pop xwa
   ret





; ---------------------------------------------------------------
; Converts XY to map memory location
; inputs hl = xy
; output xde, memory location
; ---------------------------------------------------------------
ConvertXYtoMap:
   push xbc
   
   ; multiple y by max level width
   ld xde, 0
   ld e, l
   sll MAXLEVELWIDTHFASTMUL, xde
   
   ; add x
   ld xbc, 0
   ld c, h
   add xde, xbc
   
   ; add level memory start address
   add xde, THISLEVEL
   
   pop xbc
   ret

; take hl as xy
ConvertXYToSpriteXY:
   sll 4, h
   sll 4, l
   ret
   



; ---------------------------------------------------------------
; Draws the components of the status bar
; ---------------------------------------------------------------
DrawStatus:
   calr DrawLives
   calr DrawDiamondsRemaining
   calr DrawLevelNumber
   calr DrawScore
   calr DrawTimeRemaining
   ret

DrawLives:
   push wa
   push c
   push hl
   ld c, (Lives)
   ldw hl, 0      ; position
   ld w, 4        ; palette
   ld a, ST_LIFE  ; character
DrawNextLife:
   cp c, 0
   jr z, DrawLivesEnd
   calr PrintCharTop
   inc h
   dec c
   jr DrawNextLife
DrawLivesEnd:
   ; now blank the next character
   ld a, 0
   calr PrintCharTop
   pop hl
   pop c
   pop wa
   ret

DrawDiamondsRemaining:
   push xwa
   push xbc
   push xde
   push xhl
   ldw hl, 0600h     ; position
   ld w, 2           ; palette
   ld a, ST_DIAMOND  ; char
   calr PrintCharTop
   ld w, 0
   ld xbc, 0
   ld c, (DiamondsRemaining)
   div xbc, 10
   ld de, qbc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 0800h
   calr PrintCharTop
   ld de, bc 
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 0700h
   calr PrintCharTop
   pop xhl
   pop xde
   pop xbc
   pop xwa
   ret

DrawTimeRemaining:
   push xwa
   push xbc
   push xde
   push xhl
   ldw hl, 0a00h     ; position
   ld w, 0           ; palette
   ld a, ST_TIME  ; char
   calr PrintCharTop

   ld xbc, 0
   ld c, (TimeRemaining)
   div xbc, 10
   ld de, qbc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 0d00h
   calr PrintCharTop

   ld qbc, 0
   div xbc, 10
   ld de, qbc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 0c00h
   calr PrintCharTop

   ld de, bc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 0b00h
   calr PrintCharTop

   pop xhl
   pop xde
   pop xbc
   pop xwa
   ret

DrawLevelNumber:
   push xwa
   push xbc
   push xde
   push xhl
   ld w, 0           ; palette
   ld xbc, 0
   ld c, (LevelNumber)
   div xbc, 10
   ld de, qbc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 1312h
   calr PrintCharTop
   ld de, bc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 1212h
   calr PrintCharTop
   pop xhl
   pop xde
   pop xbc
   pop xwa
   ret

DrawScore:
   push xwa
   push xbc
   push xde
   push xhl
   ld w, 0

   ld xbc, (Score)
   div xbc, 10
   ld de, qbc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 1300h
   calr PrintCharTop

   ld qbc, 0
   div xbc, 10
   ld de, qbc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 1200h
   calr PrintCharTop

   ld qbc, 0
   div xbc, 10
   ld de, qbc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 1100h
   calr PrintCharTop

   ld qbc, 0
   div xbc, 10
   ld de, qbc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 1000h
   calr PrintCharTop

   ld qbc, 0
   div xbc, 10
   ld de, qbc
   add de, ST_NUMBERS
   ld a, e
   ldw hl, 0f00h
   calr PrintCharTop
   pop xhl
   pop xde
   pop xbc
   pop xwa
   ret


; ----------------------------
; CalcPos
; ----------------------------
; inputs:
; w = pos
; a = size
; outputs:
; b = outpos
; c = outscroll
; ----------------------------

MAXSIZE	EQU 10
HALFSIZE EQU (MAXSIZE/2)

CalcPos:
	push hl

	; calc scrollarea
	ld h, 0
	ld l, a
	sub l, MAXSIZE
	jr nc, calc_pos_gt1
	jr calc_pos1
calc_pos_gt1:	
	ld h, l

calc_pos1:
	; clear outpos and outscroll
	ldw bc, 0
	
	; is position < half screen size
	ld l, w
	sub l , HALFSIZE
	jr nc, calc_pos_gt2
	ld b, w
	jr calc_pos_end
calc_pos_gt2:
	ld b, HALFSIZE
	ld w, l

	; is remainder of pos < scrollarea
	ld l, w
	sub l, h
	jr nc, calc_pos_gt3
	ld c, w
	jr calc_pos_end
calc_pos_gt3:
	ld c, h
	sub w, h

	; add remainder
	add b, w 
calc_pos_end:
	
	pop hl
	ret	




; ---------------------------------------------------------------
;
; RenderString
; Draws text on the screen (front plane)
; inputs: hl=xy, xde=pointer to null terminated text string
;         b = palette no.
;
; ---------------------------------------------------------------
RenderString:
    push xwa
    push xbc
    push xde
    push xhl
    sll 1,b
rsAgain:
    ld xwa, 0
    ld a, (xde)
    cp a, 0
    jr z, rsEnd
    cp a, ' ' ; skip spaces
    jr z, rsOver
    calr IsNumber ; cf = 1 if true
    jr c, rsNumber
    cp a, '.' ; check for period
    jr nz, rsLetter
    ld xwa, ((ch_letters_start - CHARSET) /16) + 26
    or w, b
    calr PrintChar
    add wa, 37 ; lower half of text
    inc l
    calr PrintChar
    dec l
    jr rsOver
rsLetter:
    sub a, 'A'
    add xwa, (ch_letters_start - CHARSET) /16
    or w, b
    calr PrintChar
    add wa, 37 ; lower half of text
    inc l
    calr PrintChar
    dec l
    jr rsOver
rsNumber:
    sub a, '0'
    add xwa, (ch_numbers_start - CHARSET) /16
    or w, b
    calr PrintChar
    add wa, 37 ; lower half of text
    inc l
    calr PrintChar
    dec l
rsOver:
    inc h
    inc xde
    jr rsAgain
rsEnd:
    pop xhl
    pop xde
    pop xbc
    pop xwa
    ret

; ---------------------------------------------------------------
; IsNumber
; sets the carry flag if the value if register a is between 0 - 9
;
; ---------------------------------------------------------------
Numbers db "0123456789"
IsNumber:
        push xwa
        push xde
        ld w, 0
        ld xde, Numbers
IsNum1:
        cp a, (xde)
        jr z, IsNumTrue
        inc w
        inc xde
        cp w, 10
        jr nz, IsNum1
IsNumFalse:
        rcf
        jr IsNumEnd
IsNumTrue:
        scf
IsNumEnd:
        pop xde
        pop xwa
        ret



DoTitleScreen:
	calr ClearScreen

	calr TITLE_DISPLAY

        ; ensure that the button is released before checking for exiting title
Title3:
        ld w, (rSL)
        cp w, 0
        jr nz, Title3

        ; Wait for button press
Title1:
        ld w, (rSL)
        cp w, 0
        jr z, Title1
        
        ; ensure that the button is released before exiting
Title2:
        ld w, (rSL)
        cp w, 0
        jr nz, Title2
        
        calr ClearScreen
	ret


DoDisplayGameOver:
	;first clear the screen
	calr ClearScreen

	; reset scrolling area
	ld (rS2SOX), 0
	ld (rS2SOY), 0

	; shift player off screen
	ldw hl, 0a0a0h
	calr DrawPlayer

	calr GAMEOVER_DISPLAY
        ld w, 3
	calr Sleep
	ret


; ---------------------------------------------------------------
; Start/StopTimer
; When the timer is started the VBI redraws the timer on the screen
; every second - it is stopped automatically when TimeRemaining
; reaches zero
; ---------------------------------------------------------------

StartTimer:
   ld (VBTempCounter), 0
   ld (VBStartTimer), 1
   ret

StopTimer:
   ld (VBStartTimer), 0
   ret


; ---------------------------------------------------------------
;
; vbWait
; Waits for n vertical blank interrupts
; inputs: w = number of VBIs to wait for
;
; ---------------------------------------------------------------
vbWait:
   push xwa
   push xbc
   ld (VBCOUNTER), 0
   ld c, w
vbw1:
   ld w, c

   cp (NextMoveOK), 0
   jr nz, vbw2
   cp (rSL), 0
   jr nz, vbw2
   ld (NextMoveOK), 1
vbw2:
   cp (NextMoveOK), 1
   jr nz, vbw3
   ld a, (rSL)
   cp a, 0
   jr z, vbw3
   ld (NextMove), a
vbw3:
   
   ld a, (VBCOUNTER)
   sub w, a
   jr nc, vbw1
   pop xbc
   pop xwa
   ret

; ---------------------------------------------------------------
; Sleep
; like vbWait but doesn't scan joystick
; w = sleep period (in seconds)
; ---------------------------------------------------------------
Sleep:
   push xwa
   push xbc
Sleep2:
   ld (VBCOUNTER), 0
   ld c, 50
Sleep1:
   ld b, c
   ld a, (VBCOUNTER)
   sub b, a
   jr nc, Sleep1
   dec w
   cp w, 0
   jr nz, Sleep2
   pop xbc
   pop xwa
   ret


VBlankInt:
    ; Update Watch Dog Timer to prevent CPU reset
    ld      (rWDCR),WD_CLR

    ; increment a counter
    inc (VBCOUNTER)

    ; Is the timer to be used?
    cp (VBStartTimer), 1
    jr nz, VBI_1
    
    ; don't do anything until we have reached the correct duration
    inc (VBTempCounter)
    cp (VBTempCounter), 60
    jr nz, VBI_1

    ; reset and draw new time on screen
    ld (VBTempCounter), 0
    dec (TimeRemaining)
    calr DrawTimeRemaining

    ; automatically turn ourself off when zero
    cp (TimeRemaining), 0
    jr nz, VBI_1
    ld (VBStartTimer), 0

VBI_1:
    reti

SYSTEM_CALL:
        SystemCallCode
OS_VERSION:
        OsVersionCode

    include "title.inc"
    include "gameover.inc"
    include "firefly.inc"
    include "butfly.inc"
    include "tiles.inc"
    include "levels.inc"

   IFNDEF RUNFROMMENU
   ; make the cartridge size correct
   org HEADER_OFS + 07FFFh
   db 0FFh
   ENDIF
